#!/usr/bin/env python

import json
import re
import subprocess
from dataclasses import dataclass
from datetime import date, datetime
from pathlib import Path
from typing import Literal

import polars as pl
from dataclasses_json import DataClassJsonMixin
from git import Repo

from .constants import FileTypes
from .utils import get_commit_hash_for_target_datetime, get_linguist_file_type

###############################################################################


@dataclass
class SLOCResults(DataClassJsonMixin):
    total_lines_of_code: int
    total_lines_of_comments: int
    programming_lines_of_code: int
    programming_lines_of_comments: int
    markup_lines_of_code: int
    markup_lines_of_comments: int
    prose_lines_of_code: int
    prose_lines_of_comments: int
    data_lines_of_code: int
    data_lines_of_comments: int
    unknown_lines_of_code: int
    unknown_lines_of_comments: int


def compute_sloc_metrics(
    repo_path: str | Path | Repo,
    commits_df: pl.DataFrame,
    target_datetime: str | date | datetime | None = None,
    datetime_col: Literal[
        "authored_datetime", "committed_datetime"
    ] = "authored_datetime",
) -> SLOCResults:
    # Get Repo object from path if necessary
    if isinstance(repo_path, Repo):
        repo = repo_path
    else:
        repo = Repo(repo_path)

    # Get the latest commit hexsha for the target datetime
    target_hex = get_commit_hash_for_target_datetime(
        commits_df=commits_df,
        target_datetime=target_datetime,
        datetime_col=datetime_col,
    )

    # Try to checkout the repo to that commit
    try:
        # Checkout the repo to the latest commit datetime
        repo.git.checkout(target_hex)

        # Get repo_dir from repo
        repo_dir = repo.working_dir

        # Run cloc on the repo
        pygount_output = subprocess.run(
            [
                "pygount",
                "--format=json",
                repo_dir,
            ],
            capture_output=True,
            text=True,
        )

        # Read JSON
        sloc_results = json.loads(pygount_output.stdout)

        # Iter over "files"
        total_lines_of_code = 0
        total_lines_of_comments = 0
        programming_lines_of_code = 0
        programming_lines_of_comments = 0
        markup_lines_of_code = 0
        markup_lines_of_comments = 0
        prose_lines_of_code = 0
        prose_lines_of_comments = 0
        data_lines_of_code = 0
        data_lines_of_comments = 0
        unknown_lines_of_code = 0
        unknown_lines_of_comments = 0
        for file in sloc_results["files"]:
            # Always add to total
            total_lines_of_code += file["codeCount"]
            total_lines_of_comments += file["documentationCount"]

            # Get file type
            file_type = get_linguist_file_type(file["path"])

            if file_type == FileTypes.programming.value:
                programming_lines_of_code += file["codeCount"]
                programming_lines_of_comments += file["documentationCount"]
            elif file_type == FileTypes.markup.value:
                markup_lines_of_code += file["codeCount"]
                markup_lines_of_comments += file["documentationCount"]
            elif file_type == FileTypes.prose.value:
                prose_lines_of_code += file["codeCount"]
                prose_lines_of_comments += file["documentationCount"]
            elif file_type == FileTypes.data.value:
                data_lines_of_code += file["codeCount"]
                data_lines_of_comments += file["documentationCount"]
            else:
                unknown_lines_of_code += file["codeCount"]
                unknown_lines_of_comments += file["documentationCount"]

        return SLOCResults(
            total_lines_of_code=total_lines_of_code,
            total_lines_of_comments=total_lines_of_comments,
            programming_lines_of_code=programming_lines_of_code,
            programming_lines_of_comments=programming_lines_of_comments,
            markup_lines_of_code=markup_lines_of_code,
            markup_lines_of_comments=markup_lines_of_comments,
            prose_lines_of_code=prose_lines_of_code,
            prose_lines_of_comments=prose_lines_of_comments,
            data_lines_of_code=data_lines_of_code,
            data_lines_of_comments=data_lines_of_comments,
            unknown_lines_of_code=unknown_lines_of_code,
            unknown_lines_of_comments=unknown_lines_of_comments,
        )

    finally:
        # Checkout back to HEAD
        repo.git.checkout("HEAD")


@dataclass
class RepoTagMetrics(DataClassJsonMixin):
    semver_tags_count: int
    non_semver_tags_count: int
    total_tags_count: int


def compute_tag_metrics(
    repo_path: str | Path | Repo,
    commits_df: pl.DataFrame,
    target_datetime: str | date | datetime | None = None,
    datetime_col: Literal[
        "authored_datetime", "committed_datetime"
    ] = "authored_datetime",
) -> RepoTagMetrics:
    # Get Repo object from path if necessary
    if isinstance(repo_path, Repo):
        repo = repo_path
    else:
        repo = Repo(repo_path)

    # Get the latest commit hexsha for the target datetime
    target_hex = get_commit_hash_for_target_datetime(
        commits_df=commits_df,
        target_datetime=target_datetime,
        datetime_col=datetime_col,
    )

    # Try to checkout the repo to that commit
    try:
        # Checkout the repo to the latest commit datetime
        repo.git.checkout(target_hex)

        # Get all tags
        tags = repo.tags

        # Construct semver regex
        # Direct from https://semver.org/
        # Only addition was "(v)?" to allow for "v" prefix
        semver_regex = re.compile(
            r"^(v)?(0|[1-9]\d*)\.(0|[1-9]\d*)\.(0|[1-9]\d*)(?:-((?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*)(?:\.(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*))*))?(?:\+([0-9a-zA-Z-]+(?:\.[0-9a-zA-Z-]+)*))?$"
        )

        # Count tags
        semver_tags_count = 0
        non_semver_tags_count = 0
        for tag in tags:
            if semver_regex.match(tag.name):
                semver_tags_count += 1
            else:
                non_semver_tags_count += 1

        return RepoTagMetrics(
            semver_tags_count=semver_tags_count,
            non_semver_tags_count=non_semver_tags_count,
            total_tags_count=len(tags),
        )

    finally:
        # Checkout back to HEAD
        repo.git.checkout("HEAD")



#!/usr/bin/env python

from dataclasses import dataclass
from datetime import date, datetime
from pathlib import Path
from typing import Literal

import polars as pl
from dataclasses_json import DataClassJsonMixin
from git import Repo
from tqdm import tqdm

from .constants import FileTypes
from .utils import filter_changes_to_dt_range, get_linguist_file_type

###############################################################################


@dataclass(slots=True)
class PerFileCommitDelta(DataClassJsonMixin):
    authored_datetime: datetime
    committed_datetime: datetime
    commit_hash: str
    commit_message: str
    committer_name: str | None
    committer_email: str | None
    author_name: str | None
    author_email: str | None
    filename: str
    filetype: str
    additions: int
    deletions: int
    lines_changed: int


@dataclass(slots=True)
class CommitSummary(DataClassJsonMixin):
    authored_datetime: datetime
    committed_datetime: datetime
    commit_hash: str
    commit_message: str
    committer_name: str | None
    committer_email: str | None
    author_name: str | None
    author_email: str | None
    total_files_changed: int
    total_additions: int
    total_deletions: int
    total_lines_changed: int
    programming_files_changed: int
    programming_additions: int
    programming_deletions: int
    programming_lines_changed: int
    markup_files_changed: int
    markup_additions: int
    markup_deletions: int
    markup_lines_changed: int
    prose_files_changed: int
    prose_additions: int
    prose_deletions: int
    prose_lines_changed: int
    data_files_changed: int
    data_additions: int
    data_deletions: int
    data_lines_changed: int
    unknown_files_changed: int
    unknown_additions: int
    unknown_deletions: int
    unknown_lines_changed: int


@dataclass
class ParsedCommitsResult:
    per_file_commit_deltas: pl.DataFrame
    commit_summaries: pl.DataFrame


def parse_commits(repo_path: str | Path | Repo) -> ParsedCommitsResult:
    """
    Parse the commits in a Git repository to extract per-file commit deltas
    and per-commit summaries.

    Args:
        repo_path: A string or Path to a repository,
            or a GitPython Repo object representing the repository.

    Returns:
        A ParsedCommitsResult object containing two Polars DataFrames:
            - per_file_commit_deltas: Detailed changes per file per commit.
            - commit_summaries: Summary of changes per commit.
    """
    if isinstance(repo_path, Repo):
        repo = repo_path
    else:
        repo = Repo(repo_path)

    # Get total number of commits for progress tracking
    total_commit_count = int(repo.git.rev_list("--count", "HEAD"))

    # Init storage
    per_file_commit_deltas: list[PerFileCommitDelta] = []
    per_commit_summaries: list[CommitSummary] = []

    # Iter over commits and collect data
    for commit in tqdm(
        repo.iter_commits(),
        total=total_commit_count,
        desc="Parsing commits",
        unit="commit",
        leave=False,
    ):
        authored_datetime = datetime.fromtimestamp(commit.authored_date)
        committed_datetime = datetime.fromtimestamp(commit.committed_date)
        commit_hash = commit.hexsha
        commit_message = commit.message.strip()
        committer_name = commit.committer.name if commit.committer else None
        committer_email = commit.committer.email if commit.committer else None
        author_name = commit.author.name if commit.author else None
        author_email = commit.author.email if commit.author else None

        # Init per-commit counters
        # Total
        total_files_changed = 0
        total_additions = 0
        total_deletions = 0
        total_lines_changed = 0

        # By type
        programming_files_changed = 0
        programming_additions = 0
        programming_deletions = 0
        programming_lines_changed = 0

        markup_files_changed = 0
        markup_additions = 0
        markup_deletions = 0
        markup_lines_changed = 0

        prose_files_changed = 0
        prose_additions = 0
        prose_deletions = 0
        prose_lines_changed = 0

        data_files_changed = 0
        data_additions = 0
        data_deletions = 0
        data_lines_changed = 0

        unknown_files_changed = 0
        unknown_additions = 0
        unknown_deletions = 0
        unknown_lines_changed = 0

        # Process diffs
        for filename, file_stats in commit.stats.files.items():
            # Get the file type
            filetype = get_linguist_file_type(Path(filename).name)

            # file_stats contains "insertions", "deletions", and "lines"
            additions = file_stats["insertions"]
            deletions = file_stats["deletions"]
            lines_changed = file_stats["lines"]

            # Update counts
            total_additions += additions
            total_deletions += deletions
            total_lines_changed += lines_changed
            total_files_changed += 1

            # Update type-specific counts
            if filetype == "programming":
                programming_files_changed += 1
                programming_additions += additions
                programming_deletions += deletions
                programming_lines_changed += lines_changed
            elif filetype == "markup":
                markup_files_changed += 1
                markup_additions += additions
                markup_deletions += deletions
                markup_lines_changed += lines_changed
            elif filetype == "prose":
                prose_files_changed += 1
                prose_additions += additions
                prose_deletions += deletions
                prose_lines_changed += lines_changed
            elif filetype == "data":
                data_files_changed += 1
                data_additions += additions
                data_deletions += deletions
                data_lines_changed += lines_changed
            else:
                unknown_files_changed += 1
                unknown_additions += additions
                unknown_deletions += deletions
                unknown_lines_changed += lines_changed

            # Store per-file commit delta
            per_file_commit_deltas.append(
                PerFileCommitDelta(
                    authored_datetime=authored_datetime,
                    committed_datetime=committed_datetime,
                    commit_hash=commit_hash,
                    commit_message=commit_message,
                    committer_name=committer_name,
                    committer_email=committer_email,
                    author_name=author_name,
                    author_email=author_email,
                    filename=filename,
                    filetype=filetype,
                    additions=additions,
                    deletions=deletions,
                    lines_changed=lines_changed,
                )
            )

        # Store per-commit summary
        per_commit_summaries.append(
            CommitSummary(
                authored_datetime=authored_datetime,
                committed_datetime=committed_datetime,
                commit_hash=commit_hash,
                commit_message=commit_message,
                committer_name=committer_name,
                committer_email=committer_email,
                author_name=author_name,
                author_email=author_email,
                total_files_changed=total_files_changed,
                total_additions=total_additions,
                total_deletions=total_deletions,
                total_lines_changed=total_lines_changed,
                programming_files_changed=programming_files_changed,
                programming_additions=programming_additions,
                programming_deletions=programming_deletions,
                programming_lines_changed=programming_lines_changed,
                markup_files_changed=markup_files_changed,
                markup_additions=markup_additions,
                markup_deletions=markup_deletions,
                markup_lines_changed=markup_lines_changed,
                prose_files_changed=prose_files_changed,
                prose_additions=prose_additions,
                prose_deletions=prose_deletions,
                prose_lines_changed=prose_lines_changed,
                data_files_changed=data_files_changed,
                data_additions=data_additions,
                data_deletions=data_deletions,
                data_lines_changed=data_lines_changed,
                unknown_files_changed=unknown_files_changed,
                unknown_additions=unknown_additions,
                unknown_deletions=unknown_deletions,
                unknown_lines_changed=unknown_lines_changed,
            )
        )

    return ParsedCommitsResult(
        per_file_commit_deltas=pl.DataFrame(per_file_commit_deltas),
        commit_summaries=pl.DataFrame(per_commit_summaries),
    )


def normalize_changes_df_and_remove_bot_changes(
    changes_df: pl.DataFrame,
    contributor_name_cols: tuple[str, ...] = ("committer_name", "author_name"),
    contributor_email_cols: tuple[str, ...] = ("committer_email", "author_email"),
    bot_names: tuple[str, ...] | None = ("dependabot", "github"),
    bot_email_indicators: tuple[str, ...] | None = ("[bot]",),
) -> tuple[pl.DataFrame, int]:
    # Lower case and strip all author and committer names and emails
    for contributor_name_col in contributor_name_cols:
        changes_df = changes_df.with_columns(
            pl.col(contributor_name_col)
            .str.to_lowercase()
            .str.strip_chars()
            .alias(contributor_name_col)
        )
    for contributor_email_col in contributor_email_cols:
        changes_df = changes_df.with_columns(
            pl.col(contributor_email_col)
            .str.to_lowercase()
            .str.strip_chars()
            .alias(contributor_email_col)
        )

    # Remove bot changes
    before_drop_count = len(changes_df)
    if bot_names is not None:
        changes_df = changes_df.filter(
            changes_df["committer_name"].is_in(bot_names or []).not_()
        )
    if bot_email_indicators is not None:
        changes_df = changes_df.filter(
            changes_df["committer_name"].str.contains("[bot]", literal=True).not_()
        )

    after_drop_count = len(changes_df)
    dropped_count = before_drop_count - after_drop_count

    return changes_df, dropped_count


@dataclass
class ImportantChangeDatesResults(DataClassJsonMixin):
    total_initial_change_datetime: str | None
    total_most_recent_change_datetime: str | None
    total_most_recent_substantial_change_datetime: str | None
    programming_initial_change_datetime: str | None
    programming_most_recent_change_datetime: str | None
    programming_most_recent_substantial_change_datetime: str | None
    markup_initial_change_datetime: str | None
    markup_most_recent_change_datetime: str | None
    markup_most_recent_substantial_change_datetime: str | None
    prose_initial_change_datetime: str | None
    prose_most_recent_change_datetime: str | None
    prose_most_recent_substantial_change_datetime: str | None
    data_initial_change_datetime: str | None
    data_most_recent_change_datetime: str | None
    data_most_recent_substantial_change_datetime: str | None
    unknown_initial_change_datetime: str | None
    unknown_most_recent_change_datetime: str | None
    unknown_most_recent_substantial_change_datetime: str | None


def compute_important_change_dates(
    commits_df: pl.DataFrame,
    start_datetime: str | date | datetime | None = None,
    end_datetime: str | date | datetime | None = None,
    datetime_col: Literal[
        "authored_datetime", "committed_datetime"
    ] = "authored_datetime",
    substantial_change_threshold_quantile: float = 0.1,
) -> ImportantChangeDatesResults:
    # Parse datetimes and filter commits to range
    commits_df, _, _ = filter_changes_to_dt_range(
        changes_df=commits_df,
        start_datetime=start_datetime,
        end_datetime=end_datetime,
        datetime_col=datetime_col,
    )

    change_date_results: dict[str, str | None] = {}
    for file_subset in ["total", *[ft.value for ft in FileTypes]]:
        # Get subset of changes that are relevant to this file type
        subset_df = commits_df.filter(commits_df[f"{file_subset}_lines_changed"] > 0)
        if len(subset_df) == 0:
            change_date_results[f"{file_subset}_initial_change_datetime"] = None
            change_date_results[f"{file_subset}_most_recent_change_datetime"] = None
            change_date_results[
                f"{file_subset}_most_recent_substantial_change_datetime"
            ] = None
            continue

        # Find initial change datetime
        initial_change_datetime = subset_df[datetime_col].min()

        # Find most recent commit datetime
        most_recent_change_datetime = subset_df[datetime_col].max()

        # Find threshold of lines changed
        lines_change_target = subset_df[f"{file_subset}_lines_changed"].quantile(
            substantial_change_threshold_quantile,
            interpolation="nearest",
        )

        # Filter subset to changes with at least that many lines changed
        substantial_changes_df = subset_df.filter(
            subset_df[f"{file_subset}_lines_changed"] >= lines_change_target
        )
        # Sort substantial changes by datetime descending and get most recent
        most_recent_substantial_change_datetime: datetime = substantial_changes_df.sort(
            by=datetime_col,
            descending=True,
        )[datetime_col][0]

        # Store results
        change_date_results[f"{file_subset}_initial_change_datetime"] = (
            initial_change_datetime.isoformat()
        )
        change_date_results[f"{file_subset}_most_recent_change_datetime"] = (
            most_recent_change_datetime.isoformat()
        )
        change_date_results[
            f"{file_subset}_most_recent_substantial_change_datetime"
        ] = most_recent_substantial_change_datetime.isoformat()

    return ImportantChangeDatesResults(**change_date_results)


@dataclass
class CommitCountsResults(DataClassJsonMixin):
    total_commit_count: int
    programming_commit_count: int
    markup_commit_count: int
    prose_commit_count: int
    data_commit_count: int
    unknown_commit_count: int


def compute_commit_counts(
    commits_df: pl.DataFrame,
    start_datetime: str | date | datetime | None = None,
    end_datetime: str | date | datetime | None = None,
    datetime_col: Literal[
        "authored_datetime", "committed_datetime"
    ] = "authored_datetime",
) -> CommitCountsResults:
    # Parse datetimes and filter commits to range
    commits_df, _, _ = filter_changes_to_dt_range(
        changes_df=commits_df,
        start_datetime=start_datetime,
        end_datetime=end_datetime,
        datetime_col=datetime_col,
    )

    results = {}
    for file_subset in ["total", *[ft.value for ft in FileTypes]]:
        results[f"{file_subset}_commit_count"] = len(
            commits_df.filter(commits_df[f"{file_subset}_lines_changed"] > 0)
        )

    return CommitCountsResults(**results)



#!/usr/bin/env python

import numpy as np

###############################################################################


def _compute_gini(arr: list[int]) -> float:
    # From: https://stackoverflow.com/a/61154922
    if len(arr) == 0:
        return np.nan

    np_arr = np.array(arr)
    diffsum = 0
    for i, xi in enumerate(np_arr[:-1], 1):
        diffsum += np.sum(np.abs(xi - np_arr[i:]))
    return diffsum / (len(np_arr) ** 2 * np.mean(np_arr))



#!/usr/bin/env python

import json

from repo_statistics import analyze_repositories, analyze_repository


def test_full_analysis_and_storage_of_results() -> None:
    # Repo Path can be a local path or remote
    repo_metrics = analyze_repository(
        repo_path="https://github.com/bioio-devs/bioio",
    )

    with open("example-repo-metrics.json", "w") as f:
        json.dump(repo_metrics, f, indent=4)


def test_single_disabled_metrics() -> None:
    # It is recommended to provide a GitHub API token
    # unless you disable "platform" metrics
    analyze_repository(
        repo_path="https://github.com/bioio-devs/bioio",
        # Provide a token
        # github_token="ABC",
        # Or disable platform metrics gathering
        compute_platform_metrics=False,
    )


def test_multiple_disabled_metrics() -> None:
    # Nearly every portion of metrics can be disable independent from one another
    analyze_repository(
        repo_path="https://github.com/bioio-devs/bioio",
        compute_timeseries_metrics=True,
        compute_contributor_stability_metrics=False,
        compute_contributor_absence_factor=True,
        compute_contributor_distribution_metrics=False,
        compute_repo_linter_metrics=False,
        compute_tag_metrics=True,
        compute_platform_metrics=False,
    )


def test_time_period_filters() -> None:
    # By default, all time-periods are considered
    # However, you can provide also provide a "start_datetime" and/or "end_datetime"
    analyze_repository(
        repo_path="https://github.com/bioio-devs/bioio",
        start_datetime="2025-01-01",
        end_datetime="2026-01-01",
        compute_platform_metrics=False,
    )


def test_bot_filters() -> None:
    # We also ignore bot changes by default by looking for
    # dependabot / github / [bot] account naming in commit information
    # This can be disabled, or, changed as well
    analyze_repository(
        repo_path="https://github.com/bioio-devs/bioio",
        # Keep all bots by ignoring name checks
        bot_names=None,
        # Keep all bots by ignoring email checks
        bot_email_indicators=None,
        compute_platform_metrics=False,
    )


def test_multiple_repo_processing() -> None:
    analyze_repositories(
        repo_paths=[
            "https://github.com/bioio-devs/bioio",
            "https://github.com/bioio-devs/bioio-ome-zarr",
            "https://github.com/evamaxfield/aws-grobid",
            "https://github.com/evamaxfield/rs-graph",
            "https://github.com/evamaxfield/repo-statistics",
        ],
        cache_results_path="pytest-repo-metrics.parquet",
        cache_errors_path="pytest-repo-errors.parquet",
        batch_size=2,
        use_multithreading=True,
        n_threads=2,
        compute_platform_metrics=False,
    )

    # Should basically be a no-op if re-run
    analyze_repositories(
        repo_paths=[
            "https://github.com/bioio-devs/bioio",
            "https://github.com/bioio-devs/bioio-ome-zarr",
            "https://github.com/evamaxfield/aws-grobid",
            "https://github.com/evamaxfield/rs-graph",
            "https://github.com/evamaxfield/repo-statistics",
        ],
        cache_results_path="pytest-repo-metrics.parquet",
        cache_errors_path="pytest-repo-errors.parquet",
        batch_size=2,
        use_multithreading=True,
        n_threads=2,
        compute_platform_metrics=False,
    )



#!/usr/bin/env python


def get_heuristic_project_type(
    starsgazers_count: int,
    total_contibutors_count: int,
) -> str:
    # This is a simple heuristic classification from Nadia Eghbal's "Working in Public"
    # This formulation is based on Sean Goggins' work

    # Compute ratio
    try:
        ratio_stargazers_to_contribs = starsgazers_count / total_contibutors_count
    except ZeroDivisionError:
        return "undefined"

    # Categorize
    if total_contibutors_count <= 5 and ratio_stargazers_to_contribs < 5:
        return "single-experiment"  # toy

    if total_contibutors_count > 5 and ratio_stargazers_to_contribs < 5:
        return "lab-level-tool"  # club

    if total_contibutors_count > 20 and ratio_stargazers_to_contribs >= 5:
        return "field-level-tool"  # federation

    if total_contibutors_count <= 20 and ratio_stargazers_to_contribs >= 5:
        return "specialized-instrument"  # stadium

    # "undefined" is the label for repos that do not fit into any of the above
    return "undefined"



#!/usr/bin/env python

import logging
import traceback
from collections.abc import Iterator
from dataclasses import dataclass
from datetime import date, datetime
from itertools import cycle
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Any, Literal

import polars as pl
from dataclasses_json import DataClassJsonMixin
from gh_tokens_loader import GitHubTokensCycler
from git import Repo
from timeout_function_decorator import timeout
from tqdm import tqdm

from . import (
    classification,
    commits,
    contributors,
    documentation,
    platform,
    source,
    timeseries,
    utils,
)

###############################################################################

log = logging.getLogger(__name__)

###############################################################################


@dataclass
class CoiledConfig(DataClassJsonMixin):
    keepalive: str = "5m"
    cpu: tuple[int, int] = (2, 8)
    memory: tuple[str, str] = ("4GiB", "12GiB")
    disk_size: int = 24
    n_workers: tuple[int, int] = (1, 8)
    threads_per_worker: int = 1
    spot_policy: str = "spot_with_fallback"
    extra_kwargs: dict | None = None


DEFAULT_COILED_KWARGS = CoiledConfig(
    extra_kwargs={"package_sync_conda_extras": ["git"]}
).to_dict()

###############################################################################


@dataclass
class TrackedErrorResult(DataClassJsonMixin):
    repo_path: str | Path
    err: str
    tb: str


def _analyze_repository(  # noqa: C901
    repo_path: str | Path | Repo,
    github_token: str | None = None,
    start_datetime: str | date | datetime | None = None,
    end_datetime: str | date | datetime | None = None,
    contributor_name_col: Literal["author_name", "committer_name"] = "author_name",
    datetime_col: Literal[
        "authored_datetime", "committed_datetime"
    ] = "authored_datetime",
    period_spans: tuple[str, ...] | list[str] = ("1 week", "4 weeks"),
    bot_names: tuple[str, ...] | None = ("dependabot", "github"),
    bot_email_indicators: tuple[str, ...] | None = ("[bot]",),
    substantial_change_threshold_quantile: float = 0.1,
    compute_timeseries_metrics: bool = True,
    compute_contributor_stability_metrics: bool = True,
    compute_contributor_absence_factor: bool = True,
    compute_contributor_distribution_metrics: bool = True,
    compute_repo_linter_metrics: bool = True,
    compute_sloc_metrics: bool = True,
    compute_tag_metrics: bool = True,
    compute_platform_metrics: bool = True,
) -> dict:
    # Get processed at datetime
    processed_at_dt = datetime.now()

    # Parse repo
    parsed_repo = utils.parse_repo_from_path_or_url(repo_path=repo_path)

    # Parse commits
    parsed_commit_results = commits.parse_commits(repo_path=parsed_repo.repo)
    commits_df = parsed_commit_results.commit_summaries
    per_file_commit_deltas_df = parsed_commit_results.per_file_commit_deltas

    # If less than 5 commits, return None
    if len(commits_df) < 5:
        raise ValueError(
            f"Repository {parsed_repo.owner}/{parsed_repo.name} "
            f"has less than 5 commits."
        )

    # Parse and filter changes to datetime range
    commits_df, start_datetime_dt, end_datetime_dt = utils.filter_changes_to_dt_range(
        changes_df=commits_df,
        start_datetime=start_datetime,
        end_datetime=end_datetime,
        datetime_col=datetime_col,
    )
    per_file_commit_deltas_df, _, _ = utils.filter_changes_to_dt_range(
        changes_df=per_file_commit_deltas_df,
        start_datetime=start_datetime,
        end_datetime=end_datetime,
        datetime_col=datetime_col,
    )

    # Storage for all metrics
    all_metrics: dict[str, str | None | int | float | bool] = {
        "meta_repo_owner_and_name": f"{parsed_repo.owner}/{parsed_repo.name}".lower(),
        "meta_start_datetime": start_datetime_dt.isoformat(),
        "meta_end_datetime": end_datetime_dt.isoformat(),
        "meta_contributor_name_column": contributor_name_col,
        "meta_datetime_column": datetime_col,
        "meta_processed_at": processed_at_dt.isoformat(),
        "meta_bot_names": ("---".join(bot_names) if bot_names is not None else None),
        "meta_bot_email_indicators": (
            "---".join(bot_email_indicators)
            if bot_email_indicators is not None
            else None
        ),
    }

    # Normalize and drop bot changes
    commits_df, bot_changes_count = commits.normalize_changes_df_and_remove_bot_changes(
        changes_df=commits_df,
        bot_names=bot_names,
        bot_email_indicators=bot_email_indicators,
    )
    per_file_commit_deltas_df, _ = commits.normalize_changes_df_and_remove_bot_changes(
        changes_df=per_file_commit_deltas_df,
        bot_names=bot_names,
        bot_email_indicators=bot_email_indicators,
    )
    all_metrics["bot_changes_count"] = bot_changes_count

    # Compute commit counts
    commit_count_results = commits.compute_commit_counts(
        commits_df=commits_df,
        start_datetime=start_datetime_dt,
        end_datetime=end_datetime_dt,
        datetime_col=datetime_col,
    )
    all_metrics.update(commit_count_results.to_dict())

    # Get important change dates
    important_change_date_results = commits.compute_important_change_dates(
        commits_df=commits_df,
        start_datetime=start_datetime_dt,
        end_datetime=end_datetime_dt,
        datetime_col=datetime_col,
        substantial_change_threshold_quantile=substantial_change_threshold_quantile,
    )
    all_metrics.update(important_change_date_results.to_dict())

    # Compute contributor counts
    contributor_count_results = contributors.compute_contributor_counts(
        commits_df=commits_df,
        start_datetime=start_datetime_dt,
        end_datetime=end_datetime_dt,
        contributor_name_col=contributor_name_col,
        datetime_col=datetime_col,
    )
    all_metrics.update(contributor_count_results.to_dict())

    # Compute timeseries and contributor stability metrics
    for period_span in period_spans:
        period_span_key = period_span.replace(" ", "_")

        if compute_timeseries_metrics:
            timeseries_metrics = timeseries.compute_timeseries_metrics(
                commits_df=commits_df,
                period_span=period_span,
                start_datetime=start_datetime,
                end_datetime=end_datetime,
                datetime_col=datetime_col,
            )

            for key, value in timeseries_metrics.to_dict().items():
                all_metrics[f"{period_span_key}_{key}"] = value

        if compute_contributor_stability_metrics:
            contributor_stability_metrics = (
                contributors.compute_contributor_stability_metrics(
                    commits_df=commits_df,
                    period_span=period_span,
                    start_datetime=start_datetime,
                    end_datetime=end_datetime,
                    contributor_name_col=contributor_name_col,
                    datetime_col=datetime_col,
                )
            )

            for key, value in contributor_stability_metrics.to_dict().items():
                all_metrics[f"{period_span_key}_{key}"] = value

    # Compute other contributor metrics
    if compute_contributor_absence_factor:
        contributor_absence_factor_metrics = (
            contributors.compute_contributor_absence_factor(
                commits_df=commits_df,
                start_datetime=start_datetime,
                end_datetime=end_datetime,
                contributor_name_col=contributor_name_col,
                datetime_col=datetime_col,
            )
        )

        all_metrics.update(contributor_absence_factor_metrics.to_dict())

    # Compute contributor distribution metrics
    if compute_contributor_distribution_metrics:
        contributor_distribution_metrics = (
            contributors.compute_contributor_distribution_metrics(
                per_file_commit_deltas_df=per_file_commit_deltas_df,
                start_datetime=start_datetime,
                end_datetime=end_datetime,
                contributor_name_col=contributor_name_col,
                datetime_col=datetime_col,
            )
        )

        all_metrics.update(contributor_distribution_metrics.to_dict())

    # Compute repo linter metrics
    if compute_repo_linter_metrics:
        repo_linter_results = documentation.process_with_repo_linter(
            repo_path=repo_path,
            commits_df=commits_df,
            target_datetime=end_datetime,
            datetime_col=datetime_col,
        )

        all_metrics.update(repo_linter_results.to_dict())

    # Compute SLOC metrics
    if compute_sloc_metrics:
        sloc_results = source.compute_sloc_metrics(
            repo_path=repo_path,
            commits_df=commits_df,
            target_datetime=end_datetime,
            datetime_col=datetime_col,
        )

        all_metrics.update(sloc_results.to_dict())

    # Compute tag metrics
    if compute_tag_metrics:
        tag_metrics = source.compute_tag_metrics(
            repo_path=repo_path,
            commits_df=commits_df,
            target_datetime=end_datetime,
            datetime_col=datetime_col,
        )

        all_metrics.update(tag_metrics.to_dict())

    # Compute platform metrics
    if compute_platform_metrics:
        platform_metrics = platform.compute_platform_metrics(
            repo_path=repo_path,
            github_token=github_token,
        )

        all_metrics.update(platform_metrics.to_dict())

        # Also get classification of project type
        project_type_classification = classification.get_heuristic_project_type(
            starsgazers_count=platform_metrics.stargazers_count,
            total_contibutors_count=contributor_count_results.total_contributor_count,
        )
        all_metrics["project_type_heuristic_classification"] = (
            project_type_classification
        )

    return all_metrics


def analyze_repository(
    repo_path: str | Path,
    github_token: str | None = None,
    start_datetime: str | date | datetime | None = None,
    end_datetime: str | date | datetime | None = None,
    contributor_name_col: Literal["author_name", "committer_name"] = "author_name",
    datetime_col: Literal[
        "authored_datetime", "committed_datetime"
    ] = "authored_datetime",
    period_spans: tuple[str, ...] | list[str] = ("1 week", "4 weeks"),
    bot_names: tuple[str, ...] | None = ("dependabot", "github"),
    bot_email_indicators: tuple[str, ...] | None = ("[bot]",),
    substantial_change_threshold_quantile: float = 0.1,
    compute_timeseries_metrics: bool = True,
    compute_contributor_stability_metrics: bool = True,
    compute_contributor_absence_factor: bool = True,
    compute_contributor_distribution_metrics: bool = True,
    compute_repo_linter_metrics: bool = True,
    compute_sloc_metrics: bool = True,
    compute_tag_metrics: bool = True,
    compute_platform_metrics: bool = True,
    clone_timeout_seconds: int = 60,
    analyze_timeout_seconds: int = 600,
) -> dict | TrackedErrorResult:
    # Wrap private analyze function with timeout
    @timeout(analyze_timeout_seconds)
    def _analyze_repository_with_timeout(
        **kwargs: Any,
    ) -> dict:
        return _analyze_repository(
            **kwargs,
        )

    @timeout(clone_timeout_seconds)
    def _clone_repository_with_timeout(
        repo_path: str | Path,
        to_path: str | Path,
    ) -> Repo:
        return Repo.clone_from(
            repo_path,
            to_path=to_path,
        )

    # Determine clone or path
    if isinstance(repo_path, str) and any(
        repo_path.startswith(remote_repo_prefix)
        for remote_repo_prefix in [
            "http://",
            "https://",
            "git@",
            "ssh://",
            "ftp://",
        ]
    ):
        with TemporaryDirectory() as tmpdir:
            try:
                repo = _clone_repository_with_timeout(
                    repo_path=repo_path,
                    to_path=tmpdir,
                )
            except Exception as e:
                return TrackedErrorResult(
                    repo_path=repo_path,
                    err=str(e),
                    tb=traceback.format_exc(),
                )

            try:
                metrics = _analyze_repository_with_timeout(
                    repo_path=repo,
                    github_token=github_token,
                    start_datetime=start_datetime,
                    end_datetime=end_datetime,
                    contributor_name_col=contributor_name_col,
                    datetime_col=datetime_col,
                    period_spans=period_spans,
                    bot_names=bot_names,
                    bot_email_indicators=bot_email_indicators,
                    substantial_change_threshold_quantile=substantial_change_threshold_quantile,
                    compute_timeseries_metrics=compute_timeseries_metrics,
                    compute_contributor_stability_metrics=compute_contributor_stability_metrics,
                    compute_contributor_absence_factor=compute_contributor_absence_factor,
                    compute_contributor_distribution_metrics=compute_contributor_distribution_metrics,
                    compute_repo_linter_metrics=compute_repo_linter_metrics,
                    compute_sloc_metrics=compute_sloc_metrics,
                    compute_tag_metrics=compute_tag_metrics,
                    compute_platform_metrics=compute_platform_metrics,
                )
            except Exception as e:
                return TrackedErrorResult(
                    repo_path=repo_path,
                    err=str(e),
                    tb=traceback.format_exc(),
                )

    else:
        try:
            metrics = _analyze_repository_with_timeout(
                repo_path=repo_path,
                github_token=github_token,
                start_datetime=start_datetime,
                end_datetime=end_datetime,
                contributor_name_col=contributor_name_col,
                datetime_col=datetime_col,
                period_spans=period_spans,
                bot_names=bot_names,
                bot_email_indicators=bot_email_indicators,
                substantial_change_threshold_quantile=substantial_change_threshold_quantile,
                compute_timeseries_metrics=compute_timeseries_metrics,
                compute_contributor_stability_metrics=compute_contributor_stability_metrics,
                compute_contributor_absence_factor=compute_contributor_absence_factor,
                compute_contributor_distribution_metrics=compute_contributor_distribution_metrics,
                compute_repo_linter_metrics=compute_repo_linter_metrics,
                compute_sloc_metrics=compute_sloc_metrics,
                compute_tag_metrics=compute_tag_metrics,
                compute_platform_metrics=compute_platform_metrics,
            )
        except Exception as e:
            return TrackedErrorResult(
                repo_path=repo_path,
                err=str(e),
                tb=traceback.format_exc(),
            )

    # Store the original path of the repo
    if isinstance(metrics, dict):
        metrics["repo_path"] = str(repo_path)

    return metrics


@dataclass
class AnalyzeRepositoriesResults:
    metrics_df: pl.DataFrame
    errors_df: pl.DataFrame


def _one_by_one_processing(
    batch_repo_paths: list[str | Path],
    github_token_cycler: Iterator[str | None],
    **analyze_repository_kwargs: Any,
) -> tuple[list[dict], list[TrackedErrorResult]]:
    batch_results = []
    batch_errors = []

    # Analyze batch
    for repo_path in tqdm(
        batch_repo_paths,
        desc="Analyzing repositories",
        unit="repo",
        leave=False,
    ):
        log.info(f"Analyzing repository: {repo_path}")
        result = analyze_repository(
            repo_path=repo_path,
            github_token=next(github_token_cycler),
            **analyze_repository_kwargs,
        )

        if isinstance(result, TrackedErrorResult):
            batch_errors.append(result.to_dict())
        else:
            batch_results.append(result)

    return batch_results, batch_errors


def _multiple_threads_processing(
    batch_repo_paths: list[str | Path],
    github_token_cycler: Iterator[str | None],
    n_threads: int | None = None,
    **analyze_repository_kwargs: Any,
) -> tuple[list[dict], list[TrackedErrorResult]]:
    from concurrent.futures import ThreadPoolExecutor, as_completed

    batch_results = []
    batch_errors = []

    with ThreadPoolExecutor(max_workers=n_threads) as executor:
        future_to_repo_path = {
            executor.submit(
                analyze_repository,
                repo_path=repo_path,
                github_token=next(github_token_cycler),
                **analyze_repository_kwargs,
            ): repo_path
            for repo_path in batch_repo_paths
        }

        for future in tqdm(
            as_completed(future_to_repo_path),
            total=len(batch_repo_paths),
            desc="Analyzing repositories (multi-process)",
            unit="repo",
            leave=False,
        ):
            repo_path = future_to_repo_path[future]
            try:
                result = future.result()
                if isinstance(result, TrackedErrorResult):
                    batch_errors.append(result.to_dict())
                else:
                    batch_results.append(result)

            except Exception as e:
                batch_errors.append(
                    TrackedErrorResult(
                        repo_path=repo_path,
                        err=str(e),
                        tb=traceback.format_exc(),
                    ).to_dict()
                )

    return batch_results, batch_errors


def _coiled_processing(
    batch_repo_paths: list[str | Path],
    github_token_cycler: Iterator[str | None],
    coiled_kwargs: dict | None = None,
    **analyze_repository_kwargs: Any,
) -> tuple[list[dict], list[TrackedErrorResult]]:
    try:
        import coiled
    except ImportError as e:
        raise ImportError(
            "Coiled is not installed. Please install coiled to use this feature."
        ) from e

    batch_results = []
    batch_errors = []

    # Always add git to coiled kwargs
    if coiled_kwargs is None:
        prepped_coiled_kwargs = {
            "extra_kwargs": {"package_sync_conda_extras": ["git"]},
        }
    else:
        prepped_coiled_kwargs = coiled_kwargs
        if "extra_kwargs" in prepped_coiled_kwargs:
            if "package_sync_conda_extras" in prepped_coiled_kwargs["extra_kwargs"]:
                if (
                    "git"
                    not in prepped_coiled_kwargs["extra_kwargs"][
                        "package_sync_conda_extras"
                    ]
                ):
                    prepped_coiled_kwargs["extra_kwargs"][
                        "package_sync_conda_extras"
                    ].append("git")
            else:
                prepped_coiled_kwargs["extra_kwargs"]["package_sync_conda_extras"] = [
                    "git"
                ]
        else:
            prepped_coiled_kwargs["extra_kwargs"] = {
                "package_sync_conda_extras": ["git"]
            }

    # Create coiled function
    @coiled.function(**prepped_coiled_kwargs)
    def _analyze_repository_coiled(
        repo_path: str | Path,
        github_token: str | None = None,
        **analyze_repository_kwargs: Any,
    ) -> dict | TrackedErrorResult:
        return analyze_repository(
            repo_path=repo_path,
            github_token=github_token,
            **analyze_repository_kwargs,
        )

    # Submit coiled jobs
    futures = {
        _analyze_repository_coiled.submit(
            repo_path=repo_path,
            github_token=next(github_token_cycler),
            **analyze_repository_kwargs,
        ): repo_path
        for repo_path in batch_repo_paths
    }

    for future in tqdm(
        futures,
        total=len(batch_repo_paths),
        desc="Analyzing repositories (coiled)",
        unit="repo",
        leave=False,
    ):
        repo_path = futures[future]
        try:
            result = future.result()
            if isinstance(result, TrackedErrorResult):
                batch_errors.append(result.to_dict())
            else:
                batch_results.append(result)

        except Exception as e:
            batch_errors.append(
                TrackedErrorResult(
                    repo_path=repo_path,
                    err=str(e),
                    tb=traceback.format_exc(),
                ).to_dict()
            )

    return batch_results, batch_errors


def analyze_repositories(  # noqa: C901
    repo_paths: list[str | Path],
    github_tokens: list[str | None] | str | Path | None = None,
    cache_results_path: str | Path | None = None,
    cache_errors_path: str | Path | None = None,
    ignore_cached_results: bool = False,
    batch_size: int | float = 10,
    use_multithreading: bool = False,
    n_threads: int | None = None,
    use_coiled: bool = False,
    coiled_kwargs: dict | None = None,
    **analyze_repository_kwargs: Any,
) -> AnalyzeRepositoriesResults:
    # Cannot use multiple processes and coiled at the same time
    if use_multithreading and use_coiled:
        raise ValueError("Cannot use multiple processes and coiled at the same time.")

    # Lowercase all repo paths for consistency
    repo_paths = [str(rp).lower().strip() for rp in repo_paths]

    # Check for prior cached results
    if (
        not ignore_cached_results
        and cache_results_path is not None
        and Path(cache_results_path).exists()
    ):
        log.info(f"Loading cached results from {cache_results_path}")
        previously_processed_repo_metrics = pl.read_parquet(cache_results_path)
        previously_processed_repo_paths = previously_processed_repo_metrics[
            "repo_path"
        ].to_list()
        results = previously_processed_repo_metrics.to_dicts()
    else:
        previously_processed_repo_paths = []
        results = []

    # Check for prior errored results
    if (
        not ignore_cached_results
        and cache_errors_path is not None
        and Path(cache_errors_path).exists()
    ):
        log.info(f"Loading cached errors from {cache_errors_path}")
        previously_errored_repos = pl.read_parquet(cache_errors_path)
        previously_processed_errors_repo_paths = previously_errored_repos[
            "repo_path"
        ].to_list()
        errors = previously_errored_repos.to_dicts()
    else:
        previously_processed_errors_repo_paths = []
        errors = []

    # Combine previously processed repo paths and error repo paths
    all_previously_processed_or_errored_repo_paths = {
        *previously_processed_repo_paths,
        *previously_processed_errors_repo_paths,
    }

    # Log the count of previously processed repos we are skipping
    log.info(
        f"Skipping {len(all_previously_processed_or_errored_repo_paths)} "
        f"previously processed or errored repositories."
    )

    # Remove previously processed repos from repo_paths
    to_process_repo_paths = [
        rp
        for rp in repo_paths
        if rp not in all_previously_processed_or_errored_repo_paths
    ]

    # Prepare GitHub tokens
    if github_tokens is None:
        github_tokens = [None]
    if isinstance(github_tokens, str):
        # Doesn't look like a file path
        # User provided a single token
        if not Path(github_tokens).exists():
            github_tokens = [github_tokens]

    # Check for GitHub tokens file
    if isinstance(github_tokens, (str, Path)):
        # Read with gh tokens loader
        gh_tokens_cycler = GitHubTokensCycler(github_tokens)
    elif isinstance(github_tokens, list):
        # Convert to token cycler
        gh_tokens_cycler = cycle(github_tokens)
    else:
        raise ValueError(
            "github_tokens must be None, a single str token,"
            "a list of str tokens, or a str or path to a tokens file. "
            f"Got: {type(github_tokens)}"
        )

    # Determine batch size
    if isinstance(batch_size, float):
        # Round to nearest int
        int_batch_size = round(len(to_process_repo_paths) * batch_size)
    else:
        int_batch_size = batch_size

    # Calculate total number of batches
    total_batches = (len(to_process_repo_paths) + int_batch_size - 1) // int_batch_size

    # Process in batches
    for batch_start_idx in tqdm(
        range(0, len(to_process_repo_paths), int_batch_size),
        total=total_batches,
        desc="Processing repositories in batches",
        unit="batch",
    ):
        # Get batch repo paths
        batch_repo_paths = to_process_repo_paths[
            batch_start_idx : batch_start_idx + int_batch_size
        ]

        # Use multiple threads
        if use_multithreading:
            batch_results, batch_errors = _multiple_threads_processing(
                batch_repo_paths=batch_repo_paths,
                github_token_cycler=gh_tokens_cycler,
                n_threads=n_threads,
                **analyze_repository_kwargs,
            )

        # Use coiled
        elif use_coiled:
            batch_results, batch_errors = _coiled_processing(
                batch_repo_paths=batch_repo_paths,
                github_token_cycler=gh_tokens_cycler,
                coiled_kwargs=coiled_kwargs,
                **analyze_repository_kwargs,
            )

        # Single process
        else:
            batch_results, batch_errors = _one_by_one_processing(
                batch_repo_paths=batch_repo_paths,
                github_token_cycler=gh_tokens_cycler,
                **analyze_repository_kwargs,
            )

        results.extend(batch_results)
        errors.extend(batch_errors)

        # Cache intermediate results
        if cache_results_path is not None and len(results) > 0:
            pl.DataFrame(results).write_parquet(cache_results_path)
        if cache_errors_path is not None and len(errors) > 0:
            pl.DataFrame(errors).write_parquet(cache_errors_path)

    return AnalyzeRepositoriesResults(
        metrics_df=pl.DataFrame(results),
        errors_df=pl.DataFrame(errors),
    )



#!/usr/bin/env python

import re
from dataclasses import dataclass
from datetime import date, datetime, timedelta
from functools import lru_cache
from numbers import Number
from pathlib import Path
from typing import Literal

import polars as pl
from git import Repo

from . import constants
from .data import FILE_FORMATS_TO_DTYPE_DF

###############################################################################


@lru_cache(2**14)
def get_linguist_file_type(fp: str | Path) -> str:
    """Determine the file type based on its extension using the Linguist data."""
    # Extract the filename from the path
    filename = Path(fp).name.lower()

    # Check if there is an exact filename match in the data
    filename_match = FILE_FORMATS_TO_DTYPE_DF.filter(pl.col("filename") == filename)
    if not filename_match.is_empty():
        matched_types = filename_match["type"].unique()
        if len(matched_types) == 1:
            return matched_types[0]

    # Check for extension match in the data
    extension = Path(fp).suffix.lower()
    matched = FILE_FORMATS_TO_DTYPE_DF.filter(pl.col("extension") == extension)
    if matched.is_empty():
        return constants.FileTypes.unknown.value

    matched_types = matched["type"].unique()

    # Handle single type match
    if len(matched_types) == 1:
        return matched_types[0]

    # Check if multiple types matched;
    # default to priority order: "prose", "data", "markup", "programming", "unknown"
    if len(matched_types) > 1:
        for dtype in constants.FileTypes:
            if dtype.value in matched_types:
                return dtype.value

    return constants.FileTypes.unknown.value


###############################################################################

TIMEDELTA_SIZES_SECONDS = {
    "s": 1,
    "ms": 1e-3,
    "us": 1e-6,
    "ns": 1e-9,
    "m": 60,
    "h": 60 * 60,
    "d": 60 * 60 * 24,
    "w": 7 * 60 * 60 * 24,
    "y": 365 * 60 * 60 * 24,
}

TIMEDELTA_SIZES_SECONDS_2 = {
    "second": 1,
    "millisecond": 1e-3,
    "microsecond": 1e-6,
    "nanosecond": 1e-9,
    "minute": 60,
    "hour": 60 * 60,
    "day": 60 * 60 * 24,
    "week": 7 * 60 * 60 * 24,
    "year": 365 * 60 * 60 * 24,
}
TIMEDELTA_SIZES_SECONDS_2.update(
    {k + "s": v for k, v in TIMEDELTA_SIZES_SECONDS_2.items()}
)
TIMEDELTA_SIZES_SECONDS.update(TIMEDELTA_SIZES_SECONDS_2)
TIMEDELTA_SIZES_SECONDS.update(
    {k.upper(): v for k, v in TIMEDELTA_SIZES_SECONDS.items()}
)

# Create the inversions of these as well
TIMEDELTA_SIZES_SECONDS_INV = {v: k for k, v in TIMEDELTA_SIZES_SECONDS_2.items()}

###############################################################################


def parse_timedelta(  # noqa: C901
    s: str | float | timedelta,
    default: str | Literal[False] = "seconds",
) -> timedelta:
    """Parse timedelta string to number of seconds.

    Parameters
    ----------
    s : str, float, timedelta
        String to parse, or a float representing seconds, or a timedelta object.
    default: str or False, optional
        Unit of measure if s  does not specify one. Defaults to seconds.
        Set to False to require s to explicitly specify its own unit.

    Examples
    --------
    >>> from datetime import timedelta
    >>> from dask.utils import parse_timedelta
    >>> parse_timedelta('3s')
    3
    >>> parse_timedelta('3.5 seconds')
    3.5
    >>> parse_timedelta('300ms')
    0.3
    >>> parse_timedelta(timedelta(seconds=3))  # also supports timedeltas
    3

    Notes
    -----
    This function was copied from dask.utils.parse_timedelta.
    It was modified to add support for years and linted, formatted, and typed.
    """
    if isinstance(s, timedelta):
        return s
    if isinstance(s, Number):
        s_str = str(s)
    if isinstance(s, str):
        s_str = s

    # Should be a string now, parse
    s_str = s_str.replace(" ", "")
    if not s_str[0].isdigit():
        s_str = "1" + s_str

    for i in range(len(s_str) - 1, -1, -1):
        if not s_str[i].isalpha():
            break
    index = i + 1

    prefix = s_str[:index]
    suffix = s_str[index:] or default
    if suffix is False:
        raise ValueError(f"Missing time unit: {s_str}")
    if not isinstance(suffix, str):
        raise TypeError(f"default must be str or False, got {default!r}")

    n = float(prefix)

    try:
        multiplier = TIMEDELTA_SIZES_SECONDS[suffix.lower()]
    except KeyError:
        valid_units = ", ".join(TIMEDELTA_SIZES_SECONDS.keys())
        raise KeyError(
            f"Invalid time unit: {suffix}. Valid units are: {valid_units}"
        ) from None

    result = n * multiplier
    if int(result) == result:
        result = int(result)
    return timedelta(seconds=result)


def timedelta_to_string(
    td: timedelta,
) -> str:
    """Convert a timedelta object to a human-readable string.

    Parameters
    ----------
    td : timedelta
        The timedelta object to convert.

    Returns
    -------
    str
        A human-readable string representation of the timedelta.

    Examples
    --------
    >>> from datetime import timedelta
    >>> td = timedelta(days=2)
    >>> timedelta_to_string(td)
    '2 days'
    >>> td = timedelta(hours=5)
    >>> timedelta_to_string(td)
    '5 hours'
    >>> td = timedelta(minutes=30)
    >>> timedelta_to_string(td)
    '30 minutes'
    >>> td = timedelta(days=21)
    >>> timedelta_to_string(td)
    '3 weeks'

    Notes
    -----
    This function breaks down the timedelta into its components (weeks, days,
    hours, minutes, seconds) and constructs a readable string.
    """
    # Use total seconds and find the largest fitting unit
    # That is cleanly divisble by the total seconds
    total_seconds = int(td.total_seconds())
    if total_seconds == 0:
        return "0 seconds"

    # Get sorted units by size in seconds
    sorted_units = sorted(TIMEDELTA_SIZES_SECONDS_INV.keys(), reverse=True)

    # Find the largest fitting unit
    for unit_size in sorted_units:
        if total_seconds >= unit_size and total_seconds % unit_size == 0:
            unit_name = TIMEDELTA_SIZES_SECONDS_INV[unit_size]
            count = total_seconds // unit_size
            if count == 1:
                unit_name = unit_name.rstrip("s")  # Singular form
            return f"{count} {unit_name}"

    # Fallback to seconds if no larger unit fits
    return f"{total_seconds} seconds"


def parse_datetime(s: str | date | datetime) -> datetime:
    """Parse a datetime string to a datetime object.

    Parameters
    ----------
    s : str or datetime
        String to parse, or a datetime object.

    Returns
    -------
    datetime
        Parsed datetime object.

    Examples
    --------
    >>> parse_datetime("2023-10-05T14:48:00")
    datetime.datetime(2023, 10, 5, 14, 48)
    >>> parse_datetime(datetime(2023, 10, 5, 14, 48))
    datetime.datetime(2023, 10, 5, 14, 48)

    Notes
    -----
    This function uses the built-in datetime.fromisoformat method for parsing.
    """
    if isinstance(s, datetime):
        return s
    if isinstance(s, date):
        return datetime(s.year, s.month, s.day)
    if isinstance(s, str):
        try:
            return datetime.fromisoformat(s)
        except ValueError as e:
            raise ValueError(f"Invalid datetime string: {s}") from e
    raise TypeError(f"Input must be a str or datetime, got {type(s)}")


def parse_date(s: str | datetime | date) -> date:
    """Parse a date string to a date object.

    Parameters
    ----------
    s : str, datetime, or date
        String to parse, or a datetime or date object.

    Returns
    -------
    date
        Parsed date object.

    Examples
    --------
    >>> parse_date("2023-10-05")
    datetime.date(2023, 10, 5)
    >>> parse_date(datetime(2023, 10, 5, 14, 48))
    datetime.date(2023, 10, 5)
    >>> parse_date(date(2023, 10, 5))
    datetime.date(2023, 10, 5)

    Notes
    -----
    This function uses the built-in datetime.fromisoformat method for parsing.
    """
    if isinstance(s, date) and not isinstance(s, datetime):
        return s
    if isinstance(s, datetime):
        return s.date()
    if isinstance(s, str):
        try:
            return datetime.fromisoformat(s).date()
        except ValueError as e:
            raise ValueError(f"Invalid date string: {s}") from e
    raise TypeError(f"Input must be a str, datetime, or date, got {type(s)}")


def filter_changes_to_dt_range(
    changes_df: pl.DataFrame,
    start_datetime: str | date | datetime | None = None,
    end_datetime: str | date | datetime | None = None,
    datetime_col: Literal[
        "authored_datetime", "committed_datetime"
    ] = "authored_datetime",
) -> tuple[pl.DataFrame, datetime, datetime]:
    """Filter changes DataFrame to a specified datetime range.

    Parameters
    ----------
    changes_df : pl.DataFrame
        DataFrame containing commit data or per-file-commit-deltas with
        'authored_datetime' and 'committed_datetime' columns.
    start_datetime : str, date, datetime, or None, optional
        Start of the datetime range.
        If None, uses the minimum datetime in the DataFrame.
    end_datetime : str, date, datetime, or None, optional
        End of the datetime range. If None, uses the maximum datetime in the DataFrame.
    datetime_col : str, optional
        Column to use for filtering ('authored_datetime' or 'committed_datetime').

    Returns
    -------
    pl.DataFrame
        Filtered DataFrame containing only changes within the specified datetime range.
    datetime
        The start datetime used for filtering.
    datetime
        The end datetime used for filtering.
    """
    if start_datetime is None:
        start_datetime_dt = changes_df[datetime_col].min()
    else:
        start_datetime_dt = parse_datetime(start_datetime)
    if end_datetime is None:
        end_datetime_dt = changes_df[datetime_col].max()
    else:
        end_datetime_dt = parse_datetime(end_datetime)

    # Filter changes to the specified time range
    filtered_changes_df = changes_df.filter(
        (changes_df[datetime_col] >= start_datetime_dt)
        & (changes_df[datetime_col] <= end_datetime_dt)
    )

    return filtered_changes_df, start_datetime_dt, end_datetime_dt


def get_commit_hash_for_target_datetime(
    commits_df: pl.DataFrame,
    target_datetime: str | date | datetime | None,
    datetime_col: Literal[
        "authored_datetime", "committed_datetime"
    ] = "authored_datetime",
) -> str:
    # If no target datetime is provided, return the latest commit hash
    if target_datetime is None:
        latest_commit_hexsha = commits_df.sort(datetime_col, descending=True)[
            "commit_hash"
        ][0]
        return latest_commit_hexsha

    # Find the latest commit hash up to the target datetime
    target_datetime_dt = parse_datetime(target_datetime)
    commits_up_to_target = commits_df.filter(
        commits_df[datetime_col] <= target_datetime_dt
    )
    latest_commit_hexsha = commits_up_to_target.sort(datetime_col, descending=True)[
        "commit_hash"
    ][0]
    return latest_commit_hexsha


@dataclass
class ParsedRepo:
    repo: Repo
    owner: str
    name: str


def parse_repo_from_path_or_url(
    repo_path: str | Path | Repo,
) -> ParsedRepo:
    # Get Repo object from path if necessary
    if isinstance(repo_path, Repo):
        repo = repo_path
    else:
        repo = Repo(repo_path)

    # Get the origin / remote URL
    remote_url = repo.remote().url

    # Example remote URL format:
    # git@github.com:evamaxfield/rs-graph.git
    # RegEx parse to get owner and repo name
    parsed_owner_and_name = re.match(
        r"(?:git@github\.com:|https://github\.com/)(?P<owner>[^/]+)/(?P<repo>.+)",
        remote_url,
    )
    if parsed_owner_and_name is None:
        raise ValueError(
            f"Could not parse GitHub owner and repo name from remote URL: {remote_url}"
        )

    # Extract owner and repo name
    repo_owner = parsed_owner_and_name.group("owner")
    repo_name = parsed_owner_and_name.group("repo").removesuffix(".git")

    return ParsedRepo(
        repo=repo,
        owner=repo_owner,
        name=repo_name,
    )



"""Stored data loaders."""

from __future__ import annotations

from pathlib import Path

import polars as pl
import yaml

###############################################################################
# Local storage paths

DATA_FILES_DIR = Path(__file__).parent / "files"

LINGUIST_LANGUAGES_YML = DATA_FILES_DIR / "linguist.yml"
ADDITIONAL_FORMATS_YML = DATA_FILES_DIR / "additional-formats.yml"

###############################################################################


def load_file_formats_dataframe() -> pl.DataFrame:  # noqa: C901
    """
    Read the GitHub Linguist languages YAML file into a Polars DataFrame.

    Returns:
        A Polars DataFrame containing programming language data from the
        GitHub Linguist repository.
    """
    with open(LINGUIST_LANGUAGES_YML) as open_file:
        data = yaml.safe_load(open_file)

    # Construct dataframe
    records = []
    for language, attributes in data.items():
        filenames = attributes.get("filenames", [])
        extensions = attributes.get("extensions", [])

        if not filenames and not extensions:
            records.append(
                {
                    "language": language,
                    "filename": None,
                    "extension": None,
                    "type": attributes.get("type", None),
                }
            )

        else:
            if filenames:
                if extensions:
                    for filename in filenames:
                        for extension in extensions:
                            records.append(
                                {
                                    "language": language,
                                    "filename": str(filename).lower(),
                                    "extension": str(extension).lower(),
                                    "type": attributes.get("type", None),
                                }
                            )

                else:
                    for filename in filenames:
                        records.append(
                            {
                                "language": language,
                                "filename": filename,
                                "extension": None,
                                "type": attributes.get("type", None),
                            }
                        )

            else:
                if extensions:
                    for extension in extensions:
                        records.append(
                            {
                                "language": language,
                                "filename": None,
                                "extension": extension,
                                "type": attributes.get("type", None),
                            }
                        )

    # Load the additional formats and append them to the records
    with open(ADDITIONAL_FORMATS_YML) as open_file:
        additional_data = yaml.safe_load(open_file)

    # This is much simpler structure
    # It is a list of dicts with
    for dtype, extension_list in additional_data.items():
        for extension in extension_list:
            records.append(
                {
                    "language": None,
                    "filename": None,
                    "extension": f".{extension}",
                    "type": dtype,
                }
            )

    return pl.DataFrame(records)


FILE_FORMATS_TO_DTYPE_DF = load_file_formats_dataframe()



"""Top-level package for repo-statistics."""

from importlib.metadata import PackageNotFoundError, version

try:
    __version__ = version("repo-statistics")
except PackageNotFoundError:
    __version__ = "uninstalled"

__author__ = "Eva Maxfield Brown"
__email__ = "evamaxfieldbrown@gmail.com"

from .main import (
    DEFAULT_COILED_KWARGS,
    analyze_repositories,
    analyze_repository,
)

__all__ = [
    "analyze_repositories",
    "analyze_repository",
    "DEFAULT_COILED_KWARGS",
    __version__,
    __author__,
    __email__,
]



#!/usr/bin/env python

import logging
import time
from dataclasses import dataclass
from pathlib import Path

import backoff
from dataclasses_json import DataClassJsonMixin
from ghapi.all import GhApi
from git import Repo

from .utils import parse_repo_from_path_or_url

###############################################################################

log = logging.getLogger(__name__)

###############################################################################


@dataclass
class PlatformMetrics(DataClassJsonMixin):
    primary_programming_language: str | None
    stargazers_count: int
    forks_count: int
    watchers_count: int
    open_issues_count: int


@backoff.on_exception(
    backoff.expo,
    Exception,
    max_tries=3,
)
def _request_platform_metrics_with_backoff(
    github_token: str | None,
    repo_owner: str,
    repo_name: str,
) -> PlatformMetrics:
    # Init API
    if github_token is not None:
        api = GhApi(token=github_token)
    else:
        api = GhApi()

    # Sleep to avoid rate limits
    if github_token is None:
        # Alert that we are unauthenticated and need to be careful about rate limits
        log.warning(
            "Unauthenticated GitHub API requests have string rate limits. "
            "You will likely hit these limits if you are using this library "
            "to process multiple repositories quickly. "
            "Consider providing a GitHub token to increase your rate limits."
        )

    time.sleep(0.85)

    # Request
    repo_data = api.repos.get(
        owner=repo_owner,
        repo=repo_name,
    )

    return PlatformMetrics(
        stargazers_count=repo_data["stargazers_count"],
        forks_count=repo_data["forks_count"],
        watchers_count=repo_data["watchers_count"],
        open_issues_count=repo_data["open_issues_count"],
        primary_programming_language=repo_data["language"],
    )


def compute_platform_metrics(
    repo_path: str | Path | Repo,
    github_token: str | None,
) -> PlatformMetrics:
    # Parse repo info
    parsed_repo = parse_repo_from_path_or_url(repo_path=repo_path)

    # Request platform metrics with backoff
    return _request_platform_metrics_with_backoff(
        github_token=github_token,
        repo_owner=parsed_repo.owner,
        repo_name=parsed_repo.name,
    )



#!/usr/bin/env python

from dataclasses import dataclass
from datetime import date, datetime, timedelta
from typing import Literal

import numpy as np
import polars as pl
from dataclasses_json import DataClassJsonMixin
from scipy.stats import entropy

from . import constants
from .gini import _compute_gini
from .utils import filter_changes_to_dt_range, parse_timedelta

###############################################################################


@dataclass
class ContributorCountMetrics(DataClassJsonMixin):
    total_contributor_count: int
    programming_contributor_count: int
    markup_contributor_count: int
    prose_contributor_count: int
    data_contributor_count: int
    unknown_contributor_count: int


def compute_contributor_counts(
    commits_df: pl.DataFrame,
    start_datetime: str | date | datetime | None = None,
    end_datetime: str | date | datetime | None = None,
    contributor_name_col: Literal["author_name", "committer_name"] = "author_name",
    datetime_col: Literal[
        "authored_datetime", "committed_datetime"
    ] = "authored_datetime",
) -> ContributorCountMetrics:
    # Parse datetimes and filter commits to range
    commits_df, _, _ = filter_changes_to_dt_range(
        changes_df=commits_df,
        start_datetime=start_datetime,
        end_datetime=end_datetime,
        datetime_col=datetime_col,
    )

    # Get unique contributors for each file type
    contributor_counts: dict[str, int] = {}
    for file_subset in ["total", *[ft.value for ft in constants.FileTypes]]:
        subset_df = commits_df.filter(commits_df[f"{file_subset}_lines_changed"] > 0)
        unique_contributors = subset_df[contributor_name_col].unique()
        contributor_counts[f"{file_subset}_contributor_count"] = len(
            unique_contributors
        )

    return ContributorCountMetrics(**contributor_counts)


@dataclass
class ContributorStabilityMetrics(DataClassJsonMixin):
    stable_contributors_count: int
    transient_contributors_count: int
    median_contribution_span_days: float
    mean_contribution_span_days: float
    normalized_median_contribution_span: float
    normalized_mean_contribution_span: float


def compute_contributor_stability_metrics(
    commits_df: pl.DataFrame,
    period_span: str | float | timedelta,
    start_datetime: str | date | datetime | None = None,
    end_datetime: str | date | datetime | None = None,
    contributor_name_col: Literal["author_name", "committer_name"] = "author_name",
    datetime_col: Literal[
        "authored_datetime", "committed_datetime"
    ] = "authored_datetime",
) -> ContributorStabilityMetrics:
    # Parse period span and datetimes
    td = parse_timedelta(period_span)

    # Parse datetimes and filter commits to range
    commits_df, start_datetime_dt, end_datetime_dt = filter_changes_to_dt_range(
        changes_df=commits_df,
        start_datetime=start_datetime,
        end_datetime=end_datetime,
        datetime_col=datetime_col,
    )

    # Calculate project duration in days
    project_duration = (end_datetime_dt - start_datetime_dt).days

    # Get contribution spans for each contributor
    contributor_spans = []
    contributor_stability = []
    for _, contributor_commits in commits_df.group_by(contributor_name_col):
        # Get first and last commit dates
        first_commit = contributor_commits[datetime_col].min()
        last_commit = contributor_commits[datetime_col].max()

        # Calculate contribution span in days
        span_days = (last_commit - first_commit).days
        contributor_spans.append(span_days)

        # Classify as stable/transient based on span compared to td
        if span_days >= td.days:
            contributor_stability.append("stable")
        else:
            contributor_stability.append("transient")

    # Calculate metrics
    stable_count = sum(1 for x in contributor_stability if x == "stable")
    transient_count = sum(1 for x in contributor_stability if x == "transient")

    # Calculate span statistics
    median_span = np.median(contributor_spans) if contributor_spans else 0
    mean_span = np.mean(contributor_spans) if contributor_spans else 0

    # Calculate normalized spans
    normalized_median = median_span / project_duration if project_duration > 0 else 0
    normalized_mean = mean_span / project_duration if project_duration > 0 else 0

    return ContributorStabilityMetrics(
        stable_contributors_count=stable_count,
        transient_contributors_count=transient_count,
        median_contribution_span_days=median_span,
        mean_contribution_span_days=mean_span,
        normalized_median_contribution_span=normalized_median,
        normalized_mean_contribution_span=normalized_mean,
    )


@dataclass
class ContributorAbsenceFactorMetrics(DataClassJsonMixin):
    total_contributor_absence_factor: int
    programming_contributor_absence_factor: int
    markup_contributor_absence_factor: int
    prose_contributor_absence_factor: int
    data_contributor_absence_factor: int
    unknown_contributor_absence_factor: int


def compute_contributor_absence_factor(
    commits_df: pl.DataFrame,
    start_datetime: str | date | datetime | None = None,
    end_datetime: str | date | datetime | None = None,
    contributor_name_col: Literal["author_name", "committer_name"] = "author_name",
    datetime_col: Literal[
        "authored_datetime", "committed_datetime"
    ] = "authored_datetime",
) -> ContributorAbsenceFactorMetrics:
    # Parse datetimes and filter commits to range
    commits_df, _, _ = filter_changes_to_dt_range(
        changes_df=commits_df,
        start_datetime=start_datetime,
        end_datetime=end_datetime,
        datetime_col=datetime_col,
    )

    # Create list of lines changed by file type
    all_file_subsets_lines_change_per_contrib: dict[str, list[int]] = {}
    for _, contributor_group in commits_df.group_by(contributor_name_col):
        for file_subset in ["total", *[ft.value for ft in constants.FileTypes]]:
            lines_changed_col = f"{file_subset}_lines_changed"
            if file_subset not in all_file_subsets_lines_change_per_contrib:
                all_file_subsets_lines_change_per_contrib[file_subset] = []

            all_file_subsets_lines_change_per_contrib[file_subset].append(
                contributor_group[lines_changed_col].sum()
            )

    # Sort all lists, sum, get half, then find min number of contributors to reach half
    contrib_absence_factor_per_file_subset: dict[str, int] = {}
    for (
        file_subset,
        lines_changed_per_contrib,
    ) in all_file_subsets_lines_change_per_contrib.items():
        descending_lines_changed_per_contrib = sorted(
            lines_changed_per_contrib, reverse=True
        )
        lines_changed_sum = sum(descending_lines_changed_per_contrib)
        half_lines_changed_sum = lines_changed_sum / 2
        contributors_to_half = 0
        current_total = 0
        for contributor_lines_changed in descending_lines_changed_per_contrib:
            current_total += contributor_lines_changed
            contributors_to_half += 1
            if current_total >= half_lines_changed_sum:
                break

        contrib_absence_factor_per_file_subset[
            f"{file_subset}_contributor_absence_factor"
        ] = contributors_to_half

    return ContributorAbsenceFactorMetrics(**contrib_absence_factor_per_file_subset)


@dataclass
class SingleFileSubsetContributorDistributionMetrics(DataClassJsonMixin):
    contributors_per_file_entropy: float
    contributors_per_file_gini: float
    files_per_contributor_entropy: float
    files_per_contributor_gini: float
    simple_threshold_generalist_count: int
    simple_threshold_specialist_count: int
    median_threshold_generalist_count: int
    median_threshold_specialist_count: int
    twenty_fifth_percentile_threshold_generalist_count: int
    twenty_fifth_percentile_threshold_specialist_count: int


@dataclass
class ContributorDistributionMetrics(DataClassJsonMixin):
    total_contributors_per_file_entropy: float
    total_contributors_per_file_gini: float
    total_files_per_contributor_entropy: float
    total_files_per_contributor_gini: float
    total_simple_threshold_generalist_count: int
    total_simple_threshold_specialist_count: int
    total_median_threshold_generalist_count: int
    total_median_threshold_specialist_count: int
    total_twenty_fifth_percentile_threshold_generalist_count: int
    total_twenty_fifth_percentile_threshold_specialist_count: int
    programming_contributors_per_file_entropy: float
    programming_contributors_per_file_gini: float
    programming_files_per_contributor_entropy: float
    programming_files_per_contributor_gini: float
    programming_simple_threshold_generalist_count: int
    programming_simple_threshold_specialist_count: int
    programming_median_threshold_generalist_count: int
    programming_median_threshold_specialist_count: int
    programming_twenty_fifth_percentile_threshold_generalist_count: int
    programming_twenty_fifth_percentile_threshold_specialist_count: int
    markup_contributors_per_file_entropy: float
    markup_contributors_per_file_gini: float
    markup_files_per_contributor_entropy: float
    markup_files_per_contributor_gini: float
    markup_simple_threshold_generalist_count: int
    markup_simple_threshold_specialist_count: int
    markup_median_threshold_generalist_count: int
    markup_median_threshold_specialist_count: int
    markup_twenty_fifth_percentile_threshold_generalist_count: int
    markup_twenty_fifth_percentile_threshold_specialist_count: int
    prose_contributors_per_file_entropy: float
    prose_contributors_per_file_gini: float
    prose_files_per_contributor_entropy: float
    prose_files_per_contributor_gini: float
    prose_simple_threshold_generalist_count: int
    prose_simple_threshold_specialist_count: int
    prose_median_threshold_generalist_count: int
    prose_median_threshold_specialist_count: int
    prose_twenty_fifth_percentile_threshold_generalist_count: int
    prose_twenty_fifth_percentile_threshold_specialist_count: int
    data_contributors_per_file_entropy: float
    data_contributors_per_file_gini: float
    data_files_per_contributor_entropy: float
    data_files_per_contributor_gini: float
    data_simple_threshold_generalist_count: int
    data_simple_threshold_specialist_count: int
    data_median_threshold_generalist_count: int
    data_median_threshold_specialist_count: int
    data_twenty_fifth_percentile_threshold_generalist_count: int
    data_twenty_fifth_percentile_threshold_specialist_count: int
    unknown_contributors_per_file_entropy: float
    unknown_contributors_per_file_gini: float
    unknown_files_per_contributor_entropy: float
    unknown_files_per_contributor_gini: float
    unknown_simple_threshold_generalist_count: int
    unknown_simple_threshold_specialist_count: int
    unknown_median_threshold_generalist_count: int
    unknown_median_threshold_specialist_count: int
    unknown_twenty_fifth_percentile_threshold_generalist_count: int
    unknown_twenty_fifth_percentile_threshold_specialist_count: int


def _compute_single_file_subset_contributor_distribution(
    filetype_filtered_df: pl.DataFrame,
    contributor_name_col: Literal["author_name", "committer_name"] = "author_name",
) -> SingleFileSubsetContributorDistributionMetrics:
    # Handle files per contributor
    files_per_contributor_vector = []
    for _, contributor_group in filetype_filtered_df.group_by(contributor_name_col):
        # Add the number of unique files touched by this contributor
        files_per_contributor_vector.append(contributor_group["filename"].n_unique())

    # Handle contributors per file
    contributors_per_file_vector = []
    for _, file_group in filetype_filtered_df.group_by("filename"):
        # Add the number of unique contributors who touched this file
        contributors_per_file_vector.append(file_group[contributor_name_col].n_unique())

    # Handle single contributor case
    if len(files_per_contributor_vector) == 1:
        files_per_contributor_entropy = np.nan
    else:
        # Convert to probabilities
        files_per_contributor_vector_as_prob = np.array(
            files_per_contributor_vector
        ) / sum(files_per_contributor_vector)
        files_per_contributor_entropy = entropy(
            files_per_contributor_vector_as_prob, base=2
        )

    # Handle single file case
    if len(contributors_per_file_vector) == 1:
        contributors_per_file_entropy = np.nan
    else:
        # Convert to probabilities
        contributors_per_file_vector_as_prob = np.array(
            contributors_per_file_vector
        ) / sum(contributors_per_file_vector)
        contributors_per_file_entropy = entropy(
            contributors_per_file_vector_as_prob, base=2
        )

    # Compute Gini coefficients
    contributors_per_file_gini = _compute_gini(contributors_per_file_vector)
    files_per_contributor_gini = _compute_gini(files_per_contributor_vector)

    # Count specialists and generalists
    simple_threshold_generalist_count = 0
    simple_threshold_specialist_count = 0
    median_threshold_generalist_count = 0
    median_threshold_specialist_count = 0
    twenty_fifth_percentile_threshold_generalist_count = 0
    twenty_fifth_percentile_threshold_specialist_count = 0

    # Handle no files per contributor
    if len(files_per_contributor_vector) > 0:
        # Get the median number of files changed per contributor
        twenty_fifth_percentile_files_per_contributor = np.percentile(
            files_per_contributor_vector, 25
        )
        median_files_per_contributor = np.median(files_per_contributor_vector)

        for contributor_files_count in files_per_contributor_vector:
            # Handle simple threshold
            if contributor_files_count <= 3:
                simple_threshold_specialist_count += 1
            else:
                simple_threshold_generalist_count += 1

            # Handle median threshold
            if contributor_files_count <= median_files_per_contributor:
                median_threshold_specialist_count += 1
            else:
                median_threshold_generalist_count += 1

            # Handle 25th percentile threshold
            if contributor_files_count <= twenty_fifth_percentile_files_per_contributor:
                twenty_fifth_percentile_threshold_specialist_count += 1
            else:
                twenty_fifth_percentile_threshold_generalist_count += 1

    # Compile metrics for this file subset
    return SingleFileSubsetContributorDistributionMetrics(
        contributors_per_file_entropy=contributors_per_file_entropy,
        contributors_per_file_gini=contributors_per_file_gini,
        files_per_contributor_entropy=files_per_contributor_entropy,
        files_per_contributor_gini=files_per_contributor_gini,
        simple_threshold_generalist_count=simple_threshold_generalist_count,
        simple_threshold_specialist_count=simple_threshold_specialist_count,
        median_threshold_generalist_count=median_threshold_generalist_count,
        median_threshold_specialist_count=median_threshold_specialist_count,
        twenty_fifth_percentile_threshold_generalist_count=twenty_fifth_percentile_threshold_generalist_count,
        twenty_fifth_percentile_threshold_specialist_count=twenty_fifth_percentile_threshold_specialist_count,
    )


def compute_contributor_distribution_metrics(
    per_file_commit_deltas_df: pl.DataFrame,
    start_datetime: str | date | datetime | None = None,
    end_datetime: str | date | datetime | None = None,
    contributor_name_col: Literal["author_name", "committer_name"] = "author_name",
    datetime_col: Literal[
        "authored_datetime", "committed_datetime"
    ] = "authored_datetime",
) -> ContributorDistributionMetrics:
    # Parse datetimes and filter commits to range
    per_file_commit_deltas_df, _, _ = filter_changes_to_dt_range(
        changes_df=per_file_commit_deltas_df,
        start_datetime=start_datetime,
        end_datetime=end_datetime,
        datetime_col=datetime_col,
    )

    # Make each calculation for all file types
    file_subset_metrics: dict[str, SingleFileSubsetContributorDistributionMetrics] = {}
    for file_subset in ["total", *[ft.value for ft in constants.FileTypes]]:
        # With "total" we use the whole per_file_commit_deltas_df
        # For other file types we filter down to just that file type
        if file_subset != "total":
            filetype_filtered_df = per_file_commit_deltas_df.filter(
                pl.col("filetype") == file_subset
            )
        else:
            filetype_filtered_df = per_file_commit_deltas_df

        # Store to dict
        file_subset_metrics[file_subset] = (
            _compute_single_file_subset_contributor_distribution(
                filetype_filtered_df=filetype_filtered_df,
                contributor_name_col=contributor_name_col,
            )
        )

    # Compile all file subset metrics into one dataclass
    return ContributorDistributionMetrics(
        **{
            f"{file_subset}_{metric_name}": metric_value
            for file_subset, file_subset_metrics in file_subset_metrics.items()
            for metric_name, metric_value in file_subset_metrics.to_dict().items()
        },
    )


@dataclass
class ContributorChangeMetrics(DataClassJsonMixin):
    diff_contributor_count: int
    same_contributor_count: int


def compute_contributor_change_metrics(
    commits_df: pl.DataFrame,
    start_datetime: str | date | datetime | None = None,
    end_datetime: str | date | datetime | None = None,
    datetime_col: Literal[
        "authored_datetime", "committed_datetime"
    ] = "authored_datetime",
    contributor_name_col: Literal["author_name", "committer_name"] = "author_name",
) -> ContributorChangeMetrics:
    # Parse datetimes and filter commits to range
    commits_df, _, _ = filter_changes_to_dt_range(
        changes_df=commits_df,
        start_datetime=start_datetime,
        end_datetime=end_datetime,
        datetime_col=datetime_col,
    )

    # Calculate commit count thresholds
    total_commits = len(commits_df)
    commit_threshold = int(total_commits * 0.2)

    # Get the set of contributors in the first 20% of commits
    initial_commits = commits_df.head(commit_threshold)

    # Ensure that we at least have 3 commits
    if len(initial_commits) < 3:
        initial_commits = commits_df.head(3)

    # Get contribs in the first 20%
    initial_contributors = set(initial_commits[contributor_name_col])

    # Get the set of contributors in the last 20% of commits
    most_recent_commits = commits_df.tail(commit_threshold)

    # Ensure that we at least have 3 commits
    if len(most_recent_commits) < 3:
        most_recent_commits = commits_df.tail(3)

    # Get contribs in the last 20%
    most_recent_contributors = set(most_recent_commits[contributor_name_col])

    # Get the difference
    return ContributorChangeMetrics(
        diff=len(initial_contributors.difference(most_recent_contributors)),
        same=len(initial_contributors.intersection(most_recent_contributors)),
    )



#!/usr/bin/env python

import json
import re
from dataclasses import dataclass
from datetime import date, datetime
from pathlib import Path
from typing import Literal

import polars as pl
from dataclasses_json import DataClassJsonMixin
from git import Repo

from .data import DATA_FILES_DIR
from .utils import get_commit_hash_for_target_datetime

###############################################################################

REPO_LINTER_RULESET_PATH = DATA_FILES_DIR / "repo-linter-ruleset.json"

###############################################################################


def _process_repo_linter_file_existence_rule(
    repo: Repo,
    rule_details: dict,
    check_dir: bool = False,
) -> bool:
    # Get repo_dir from repo
    repo_dir = repo.working_dir

    # Check for certain parameters
    allow_directories = rule_details.get("dirs", False)
    ignore_case = rule_details.get("nocase", False)

    # Get all of the glob patterns to check for
    glob_patterns = rule_details["globsAny"]

    # Iterate over all of the glob patterns and store the boolean values in a list
    glob_results = []
    for glob_pattern in glob_patterns:
        # Run the glob pattern
        if ignore_case:
            glob_result = list(Path(repo_dir).glob(glob_pattern, case_sensitive=False))
        else:
            glob_result = list(Path(repo_dir).glob(glob_pattern))

        # If we are not allowing directories, filter out directories
        if not allow_directories:
            glob_result = [fp for fp in glob_result if not fp.is_dir()]

        # Check dir if we are checking for directories
        if check_dir:
            glob_result = [fp for fp in glob_result if fp.is_dir()]

        # Check if there are any files present
        glob_results.append(len(glob_result) > 0)

    # Check if any of the glob results are True
    return len(glob_results) >= 1 and any(glob_results)


def _process_repo_linter_file_contents_rule(
    repo: Repo,
    rule_details: dict,
) -> bool:
    # Get repo_dir from repo
    repo_dir = repo.working_dir

    # Check for certain parameters
    regex_content = rule_details["content"]
    ignore_case = rule_details.get("nocase", False)
    regex_flags = rule_details.get("flags", "")

    # The only flag that is possible present is "ignore case" so just check for that
    if "i" in regex_flags:
        ignore_case = True

    # Globs to parse
    glob_patterns = rule_details["globsAll"]

    # Iterate over all of the glob patterns and store the boolean values in a list
    glob_results = []
    for glob_pattern in glob_patterns:
        # Run the glob pattern
        glob_result = list(Path(repo_dir).glob(glob_pattern))

        # Check if there are any files present
        for fp in glob_result:
            with open(fp) as open_f:
                file_contents = open_f.read()

                # Check if we should ignore case
                if ignore_case:
                    if re.search(regex_content, file_contents, flags=re.IGNORECASE):
                        glob_results.append(True)
                    else:
                        glob_results.append(False)
                else:
                    if re.search(regex_content, file_contents):
                        glob_results.append(True)
                    else:
                        glob_results.append(False)

    # Check if all of the glob results are True
    return len(glob_results) >= 1 and all(glob_results)


def _process_repo_linter_file_type_exclusion_rule(
    repo: Repo,
    rule_details: dict,
) -> bool:
    # Get repo_dir from repo
    repo_dir = repo.working_dir

    # Iter over all "type" globs
    glob_results = []
    for glob_pattern in rule_details["type"]:
        # Run the glob pattern
        glob_result = list(Path(repo_dir).glob(glob_pattern))

        # Check if there are any files present
        glob_results.append(len(glob_result) > 0)

    # Check if any of the glob results are True
    return len(glob_results) == 0 or not any(glob_results)


def _process_repo_linter_rule(
    repo: Repo,
    rule_type: str,
    rule_details: dict,
) -> bool:
    # Process based on rule type
    if rule_type == "file-existence":
        return _process_repo_linter_file_existence_rule(
            repo=repo,
            rule_details=rule_details,
        )

    if rule_type == "directory-existence":
        return _process_repo_linter_file_existence_rule(
            repo=repo,
            rule_details=rule_details,
            check_dir=True,
        )

    if rule_type == "file-contents":
        return _process_repo_linter_file_contents_rule(
            repo=repo,
            rule_details=rule_details,
        )

    if rule_type == "file-type-exclusion":
        return _process_repo_linter_file_type_exclusion_rule(
            repo=repo,
            rule_details=rule_details,
        )

    raise ValueError(f"Unknown rule type: {rule_type}")


@dataclass
class RepoLinterResults(DataClassJsonMixin):
    documentation_checks_passed_count: int
    license_file_exists: bool
    readme_file_exists: bool
    contributing_file_exists: bool
    code_of_conduct_file_exists: bool
    changelog_file_exists: bool
    security_file_exists: bool
    support_file_exists: bool
    readme_references_license: bool
    binaries_not_present: bool
    test_directory_exists: bool
    integrates_with_ci: bool
    code_of_conduct_file_contains_email: bool
    github_issue_template_exists: bool
    github_pull_request_template_exists: bool


def process_with_repo_linter(
    repo_path: str | Path | Repo,
    commits_df: pl.DataFrame,
    target_datetime: str | date | datetime | None = None,
    datetime_col: Literal[
        "authored_datetime", "committed_datetime"
    ] = "authored_datetime",
) -> RepoLinterResults:
    # Get Repo object from path if necessary
    if isinstance(repo_path, Repo):
        repo = repo_path
    else:
        repo = Repo(repo_path)

    # Get the latest commit hexsha for the target datetime
    target_hex = get_commit_hash_for_target_datetime(
        commits_df=commits_df,
        target_datetime=target_datetime,
        datetime_col=datetime_col,
    )

    # Try to checkout the repo to that commit
    try:
        # Checkout the repo to the latest commit datetime
        repo.git.checkout(target_hex)

        # Load the ruleset
        with open(REPO_LINTER_RULESET_PATH) as open_file:
            repo_linter_full_ruleset_data = json.load(open_file)

        # Get the rules
        repo_linter_rules = repo_linter_full_ruleset_data["rules"]

        # Process each rule
        rule_results = {}
        for rule_name, rule_full_details in repo_linter_rules.items():
            # Unpack full details to get rule type and rule details
            rule_type = rule_full_details["rule"]["type"]
            rule_details = rule_full_details["rule"]["options"]

            # Process rule
            rule_result = _process_repo_linter_rule(
                repo=repo,
                rule_type=rule_type,
                rule_details=rule_details,
            )

            # Add to rule results
            rule_results[rule_name] = rule_result

        return RepoLinterResults(
            documentation_checks_passed_count=sum(
                int(value) for value in rule_results.values()
            ),
            **{
                rule_name.replace("-", "_"): rule_result
                for rule_name, rule_result in rule_results.items()
            },
        )

    finally:
        # Checkout back to HEAD
        repo.git.checkout("HEAD")



#!/usr/bin/env python

import math
from dataclasses import dataclass
from datetime import date, datetime, timedelta
from typing import Literal

import numpy as np
import polars as pl
from dataclasses_json import DataClassJsonMixin
from scipy.stats import entropy, variation
from tqdm import tqdm

from . import constants
from .gini import _compute_gini
from .utils import filter_changes_to_dt_range, parse_timedelta, timedelta_to_string

###############################################################################


@dataclass
class ChangePeriodResults(DataClassJsonMixin):
    period_span: str
    start_datetime: str
    end_datetime: str
    datetime_column: str
    total_changed_binary: list[int]
    total_lines_changed_count: list[int]
    programming_changed_binary: list[int]
    programming_lines_changed_count: list[int]
    markup_changed_binary: list[int]
    markup_lines_changed_count: list[int]
    prose_changed_binary: list[int]
    prose_lines_changed_count: list[int]
    data_changed_binary: list[int]
    data_lines_changed_count: list[int]
    unknown_changed_binary: list[int]
    unknown_lines_changed_count: list[int]


def get_periods_changed(
    commits_df: pl.DataFrame,
    period_span: str | float | timedelta,
    start_datetime: str | date | datetime | None = None,
    end_datetime: str | date | datetime | None = None,
    datetime_col: Literal[
        "authored_datetime",
        "committed_datetime",
    ] = "authored_datetime",
) -> ChangePeriodResults:
    # Parse period span and datetimes
    td = parse_timedelta(period_span)

    # Parse datetimes and filter commits to range
    commits_df, start_datetime_dt, end_datetime_dt = filter_changes_to_dt_range(
        changes_df=commits_df,
        start_datetime=start_datetime,
        end_datetime=end_datetime,
        datetime_col=datetime_col,
    )

    # Calculate total periods
    change_duration = end_datetime_dt - start_datetime_dt
    n_tds = math.ceil(change_duration / td)

    # Iterate over periods and record binary or lines changed count
    current_start_dt = start_datetime_dt
    results: dict[str, list[int]] = {}
    for _ in tqdm(range(n_tds), desc="Processing change periods", leave=False):
        # Get the subset of commits in this period
        commit_subset = commits_df.filter(
            pl.col(datetime_col).is_between(
                current_start_dt,
                current_start_dt + td,
                closed="left",
            )
        )

        # Update changes (binary and lines) for this period
        def _get_binary_and_count_from_file_subset(
            commit_subset: pl.DataFrame, file_subset: str
        ) -> tuple[int, int]:
            lines_changed_col = f"{file_subset}_lines_changed"
            return (
                int(len(commit_subset.filter(pl.col(lines_changed_col) > 0)) > 0),
                commit_subset[lines_changed_col].sum(),
            )

        # Iter over each file subset
        for file_subset in ["total", *[ft.value for ft in constants.FileTypes]]:
            # Process period
            binary, count = _get_binary_and_count_from_file_subset(
                commit_subset,
                file_subset,
            )

            # Check if list already exists at key
            changed_binary_key = f"{file_subset}_changed_binary"
            lines_changed_count_key = f"{file_subset}_lines_changed_count"
            if changed_binary_key not in results:
                results[changed_binary_key] = []
            if lines_changed_count_key not in results:
                results[lines_changed_count_key] = []

            # Append to the lists
            results[changed_binary_key].append(binary)
            results[lines_changed_count_key].append(count)

        # Increment
        current_start_dt += td

    return ChangePeriodResults(
        period_span=timedelta_to_string(td),
        start_datetime=start_datetime_dt.isoformat(),
        end_datetime=end_datetime_dt.isoformat(),
        datetime_column=datetime_col,
        **results,
    )


@dataclass
class ChangeSpanResults(DataClassJsonMixin):
    did_change_spans: list[int]
    did_not_change_spans: list[int]


def get_change_spans(
    periods_changed: list[int],
) -> ChangeSpanResults:
    # Iter over list of periods changed and count periods between changes
    did_change_spans = []
    did_not_change_spans = []
    in_change_span = False
    did_change_current_periods = 0
    did_not_change_current_periods = 0
    for period in periods_changed:
        if not in_change_span and period == 0:
            did_not_change_current_periods += 1
        if not in_change_span and period == 1:
            did_not_change_spans.append(did_not_change_current_periods)
            did_not_change_current_periods = 0
            in_change_span = True
            did_change_current_periods += 1
        if in_change_span and period == 0:
            did_change_spans.append(did_change_current_periods)
            did_change_current_periods = 0
            in_change_span = False
            did_not_change_current_periods += 1
        if in_change_span and period == 1:
            did_change_current_periods += 1

    # Remove extra zero in either of the lists
    did_change_spans = [span for span in did_change_spans if span != 0]
    did_not_change_spans = [span for span in did_not_change_spans if span != 0]

    return ChangeSpanResults(
        did_change_spans=did_change_spans,
        did_not_change_spans=did_not_change_spans,
    )


@dataclass
class TimeseriesMetrics(DataClassJsonMixin):
    # Change existance metrics
    total_changed_binary_entropy: float
    total_changed_binary_gini: float
    total_changed_binary_variation: float
    total_changed_binary_frac: float
    programming_changed_binary_entropy: float
    programming_changed_binary_gini: float
    programming_changed_binary_variation: float
    programming_changed_binary_frac: float
    markup_changed_binary_entropy: float
    markup_changed_binary_gini: float
    markup_changed_binary_variation: float
    markup_changed_binary_frac: float
    prose_changed_binary_entropy: float
    prose_changed_binary_gini: float
    prose_changed_binary_variation: float
    prose_changed_binary_frac: float
    data_changed_binary_entropy: float
    data_changed_binary_gini: float
    data_changed_binary_variation: float
    data_changed_binary_frac: float
    unknown_changed_binary_entropy: float
    unknown_changed_binary_gini: float
    unknown_changed_binary_variation: float
    unknown_changed_binary_frac: float
    # Lines changed count metrics
    total_lines_changed_count_entropy: float
    total_lines_changed_count_gini: float
    total_lines_changed_count_variation: float
    programming_lines_changed_count_entropy: float
    programming_lines_changed_count_gini: float
    programming_lines_changed_count_variation: float
    markup_lines_changed_count_entropy: float
    markup_lines_changed_count_gini: float
    markup_lines_changed_count_variation: float
    prose_lines_changed_count_entropy: float
    prose_lines_changed_count_gini: float
    prose_lines_changed_count_variation: float
    data_lines_changed_count_entropy: float
    data_lines_changed_count_gini: float
    data_lines_changed_count_variation: float
    unknown_lines_changed_count_entropy: float
    unknown_lines_changed_count_gini: float
    unknown_lines_changed_count_variation: float
    # Change span metrics
    total_did_change_median_span: int
    total_did_change_mean_span: float
    total_did_change_std_span: float
    total_did_not_change_median_span: int
    total_did_not_change_mean_span: float
    total_did_not_change_std_span: float
    programming_did_change_median_span: int
    programming_did_change_mean_span: float
    programming_did_change_std_span: float
    programming_did_not_change_median_span: int
    programming_did_not_change_mean_span: float
    programming_did_not_change_std_span: float
    markup_did_change_median_span: int
    markup_did_change_mean_span: float
    markup_did_change_std_span: float
    markup_did_not_change_median_span: int
    markup_did_not_change_mean_span: float
    markup_did_not_change_std_span: float
    prose_did_change_median_span: int
    prose_did_change_mean_span: float
    prose_did_change_std_span: float
    prose_did_not_change_median_span: int
    prose_did_not_change_mean_span: float
    prose_did_not_change_std_span: float
    data_did_change_median_span: int
    data_did_change_mean_span: float
    data_did_change_std_span: float
    data_did_not_change_median_span: int
    data_did_not_change_mean_span: float
    data_did_not_change_std_span: float
    unknown_did_change_median_span: int
    unknown_did_change_mean_span: float
    unknown_did_change_std_span: float
    unknown_did_not_change_median_span: int
    unknown_did_not_change_mean_span: float
    unknown_did_not_change_std_span: float


def _compute_entropy(arr: list[int]) -> float:
    arr_sum = np.sum(arr)
    if arr_sum == 0:
        return np.nan

    return entropy(
        np.array(arr) / arr_sum,
        base=2,
    )


def _compute_frac(arr: list[int]) -> float:
    return np.sum(arr) / len(arr)


def _compute_metrics_from_periods_change_results(
    periods_changed_results: ChangePeriodResults,
) -> dict[str, int | float]:
    # Iter over all non-metadata items returned in periods_changed_results
    # Compute single metrics in-place of arrays
    period_and_span_metrics: dict[str, int | float] = {}
    for period_key, metadata_or_arr in periods_changed_results.to_dict().items():
        # Ignore metadata fields
        if period_key in (
            "period_span",
            "start_datetime",
            "end_datetime",
            "datetime_column",
        ):
            continue

        # All metadata should not be filtered out
        arr = metadata_or_arr
        assert isinstance(arr, list)

        # Compute
        period_and_span_metrics[f"{period_key}_entropy"] = _compute_entropy(arr)
        period_and_span_metrics[f"{period_key}_gini"] = _compute_gini(arr)
        period_and_span_metrics[f"{period_key}_variation"] = variation(arr)
        if "binary" in period_key:
            period_and_span_metrics[f"{period_key}_frac"] = _compute_frac(arr)

            # All "binary" keys follow pattern of
            # <file_subset>_changed_binary
            file_subset = period_key.replace("_changed_binary", "")

            # Get spans
            span_results = get_change_spans(arr)
            for span_reduction_func in [
                np.median,
                np.mean,
                np.std,
            ]:
                # Did change spans
                did_change_span_key = (
                    f"{file_subset}_did_change_" f"{span_reduction_func.__name__}_span"
                )
                if len(span_results.did_change_spans) > 0:
                    period_and_span_metrics[did_change_span_key] = span_reduction_func(
                        span_results.did_change_spans
                    )
                else:
                    period_and_span_metrics[did_change_span_key] = float("nan")

                # Did not change spans
                did_not_change_span_key = (
                    f"{file_subset}_did_not_change_"
                    f"{span_reduction_func.__name__}_span"
                )
                if len(span_results.did_not_change_spans) > 0:
                    period_and_span_metrics[did_not_change_span_key] = (
                        span_reduction_func(span_results.did_not_change_spans)
                    )
                else:
                    period_and_span_metrics[did_not_change_span_key] = float("nan")

    return period_and_span_metrics


def compute_timeseries_metrics(
    commits_df: pl.DataFrame,
    period_span: str | float | timedelta,
    start_datetime: str | date | datetime | None = None,
    end_datetime: str | date | datetime | None = None,
    datetime_col: Literal[
        "authored_datetime", "committed_datetime"
    ] = "authored_datetime",
) -> TimeseriesMetrics:
    # Parse period span and datetimes
    td = parse_timedelta(period_span)

    # Parse datetimes and filter commits to range
    commits_df, start_datetime_dt, end_datetime_dt = filter_changes_to_dt_range(
        changes_df=commits_df,
        start_datetime=start_datetime,
        end_datetime=end_datetime,
        datetime_col=datetime_col,
    )

    # Get periods changed
    periods_changed_results = get_periods_changed(
        commits_df=commits_df,
        period_span=td,
        start_datetime=start_datetime_dt,
        end_datetime=end_datetime_dt,
        datetime_col=datetime_col,
    )

    # Compute metrics from periods changed results
    # 1. Entropy, gini, and variation of binary and lines changed count
    # 2. Fraction of periods with changes (for binary only)
    # 3. Change spans (for binary only)
    #    - Median, mean, std of spans with changes
    #    - Median, mean, std of spans without changes
    period_and_span_metrics = _compute_metrics_from_periods_change_results(
        periods_changed_results,
    )

    return TimeseriesMetrics(
        **period_and_span_metrics,  # type: ignore
    )
